//****************************************************************************
// 6510/6502 verilog cpu. with asynchonous memory interface
//
// (C) 2013, Ken Chamberlain <kenneth.r.chamberlain@comcast.net>
//
// Feel free to use this code in any project (commercial or not), as long as you
// keep this message, and the copyright notice. This code is provided "as is", 
// without any warranties of any kind. 
// 
//****************************************************************************

//****************************************************************************
//X6510A

//The 32 bit read port is used for reading instructions, vector addresses
//or external operands in page 2 or higher.

//The 8 bit read port is used for io operand read operations.

//The 8 bit write port is used for all write operations (io or memory).
//Stacking operations are not written to external memory.

//Almost all instructions take fewer clock cycles than the original processor.
//mos6502 clock cycle counts, X6510A cycle counts are shown below.
//There is also an extra state that will be entered to load more instruction
//bytes if necessary at the end of normal execution of an instruction.
//This state is not entered very often that I can tell from my testing.
//If you lower the ir queue size it wil start happenning.

//CPU states (S_EXECUTE_INSTn) are the starting instruction execute states
//for all instructions based on if it is a 1, 2 or 3 byte instruction. Some
//previous states were used to load and decode the instruction.
//To calculate cycle counts start counting at state S_EXECUTE_INSTn until you
//get back to a state S_EXECUTE_INSTn. Do not count state S_EXECUTE_INSTn
//again as that is now processing the next instruction. An extra cycle
//may be added to some instructions depending on how full the instruction
//queue is at the time.

//Unimplemented instructions will halt the processor.

//BCD arithmatic is supported but takes 1 additional clock cycle to complete.

//The processor can be stopped at any time with a high level on pause or
//low levels on vic_req or dma_req. When pause goes low the current memory
//cycle is completed.

//Default external memory accesses are instruction fetches, if there is room for
//another 4 bytes in the ir queue they will be loaded into the ir queue
//If the instruction queue cannot take in 4 new bytes the next instruction
//fetch address is still read but the data is disregarded.

//Two types of dma are supported, burst type which is used by VIC requests
//for video and single cycle requests that can be used by some other device.
//Burst requests are acknowledged by vic_ack going low and will always occur
//on the next memory cycle of the request. A single cycle dma request is
//acknowleged by dma_ack. This is a lower priority dma and may be held off
//by a burst request at the same time or already in progress. The burst
//requests can never write to memory currently.

//The RDY input can be used to stall the processor during any memory cycle.
//Unlike the original 65XX processors, writes will be stalled too.
//If a dma request is acknowledged before RDY goes high that memory cycle
//is aborted and will be restarted again as soon as cpu_ack is low again.
//RDY logic needs to take this into account by reseting itself if necessary.
//When the cpu is not the master the dma address, r/w and wb lines will be
//continually passed on to the memory busses so the dma master will have to
//include the RDY line in its driving logic if slow access may be required.
//So any slow access device that has its cycle aborted (RDY never set high)
//must be able to handle a restart of that access at a later time without
//causing a problem.

//Code cannot be executed out of the stack page unless the instructions are
//written there by normal store commands. Pushing a routine onto the stack
//and then trying to execute it will not work.

//`define N6502  //remove the C64 memory map register...
//`define debug  //video display of registers, pc breakpoint & single stepping.
//`define debug_cycles

//              __    __    __    __    __    _     __    __    __    __    __    __    __    __
//     clk  |__|  |__|  |__|  |__|  |__|  |__| ..._|  |__|  |__|  |__|  |__|  |__|  |__|  |__|
//
// cpu_a(al)   X A0  X A1  X A2  X A3  X V1  X ... X Vm  X Vn  X XX  X A4  X A5  X A6  X A7  X
//
//  cpu_rd     X i0  X i1  X i2  X i3  X ... X XX  X XX  X XX  X XX  X i4  X i5  X i6  X
//
// cpu_ia(ial) X B0  X B1  X B2  X B3  X XX  X ... X XX  X XX  X D0  X B4  X B5  X B6  X B7  X
//             _______       __________________   ______________       _________________________
//  cpu_rw            |_____|                  ...              |_____|
//
//  cpu_pr     X O0  X XX  X O2  X O3  X ... X XX  X XX  X XX  X XX  X O4  X O5  X O6  X
//
//  cpu_wb     X XX  X w1  X XX  X XX  X XX  X ... X XX  X XX  X wd  X XX  X XX  X XX  X XX  X
//             ____________________                         ____________________________________
// vic_req*                        |___________..._________|
//             _________________________                         _______________________________
// vic_ack*                             |______...______________|
//             ____________________________                            _________________________
// dma_req*                                |___...____________________|
//             ________________________________   ______________       _________________________
// dma_ack*                                    ...              |_____|
//                                       ______   ____________________
// cpu_ack*  XX_________________________|      ...                    |_________________________
//              __    __    __    __    __    _     __    __    __    __    __    __    __    __
//     clk  |__|  |__|  |__|  |__|  |__|  |__| ..._|  |__|  |__|  |__|  |__|  |__|  |__|  |__|
//                                         

//****************************************************************************
module X6510A(clk,      // system memory clock
              RDY,      // active high memory cycle complete signal
              cpu_a,    // [15:0] cpu/vic/dma address for accessing cpu_rd[31:16]
              cpu_al,   // [15:2] cpu/vic/dma address for accessing cpu_rd[15:0]
              cpu_rd,   // [31:0] instruction, vector or operand fetching data bus
              pRead,    // indicates a peripheral read and data should be read off the cpu_pr bus
              cpu_pr,   // [7:0] data operand read data (IO read)
              cpu_wb,   // [7:0] data operand write data (memory or io)
              cpu_rw,   // read(H) / write(L) control signal for external accesses
              cpu_wbyte,// [3:0] write{L} byte selects - only one of these will ever go low at a time.
              vic_req,  // burst read dma request
              vic_ack,  // burst read dma acknowledge - 1 cycle delay from request
              vic_a,    // [15:1] burst read source address => cpu_a, cpu_al
              dma_req,  // single cycle dma access request 
                        // *** must be released on the next clk cycle after dma_ack is asserted ***
              dma_ack,  // single cycle dma acknowledge - next cycle free from burst request will be used
              dma_a,    // [15:0] single cycle dma source address => cpu_a, cpu_al, cpu_sa, cpu_sal
                        // currently access to the internal stack and zp is not available - 
                        // it could be made available through the write byte port.
                        // External ZP is kept in sync with internal memory but stack operations are not.
              dma_dw,   // [7:0] single cycle dma data write source => cpu_wb
              dma_rw,   // single cycle dma read(H) / write(L) control signal => cpu_rw
              nmi_req,  // negative edge non-maskable interrupt input
              irq,      // low level maskable interrupt input
              resetIn,  // low level cpu reset input
              cpu_pause // low level cpu pause input
             
`ifndef N6502
              , cpu_hiram  // C64 memory map control signal out
              , cpu_loram  // C64 memory map control signal out
              , cpu_charen // C64 memory map control signal out
`endif             
`ifdef debug
              , debug_add // [4:0] video select of what to be displayed on its current raster
              , debug_Dout// [47:0] 8 characters x 6 bits data value to be displayed - supports 64 different characters
              , debug_sel // [2:0] selects what dataset is to be displayed, registers, pc trace etc...
              , breakTrig // low level output that indicates the pc breakpoint was reached (should cause pause to occur)
              , break_pc  // [15:0] pc compare address for program breakpoint
              , cpu_ack   // this is low when the cpu has control of the memory bus - needed for single stepping
//              , state     // cpu state machine value
`endif
             );
//****************************************************************************
input clk, RDY;

input vic_req;
input [15:1] vic_a;
input nmi_req, irq, resetIn, cpu_pause;
input [31:0] cpu_rd;
input [7:0] cpu_pr;
input pRead;

output [15:0] cpu_a;
output [15:2] cpu_al;
output [7:0] cpu_wb;
output [3:0] cpu_wbyte;

output vic_ack, cpu_rw;

input dma_req, dma_rw;
input [15:0] dma_a;
input [7:0] dma_dw;
output dma_ack;

`ifdef debug
input [15:0] break_pc;
input [4:0] debug_add;
input [2:0] debug_sel;
output [47:0] debug_Dout;
//output [3:0] state;
output breakTrig, cpu_ack;
`endif

//****************************************************************************
//acknowledge signals for bus mastership - active low
reg vic_ack = H;
reg dma_ack = H;
reg cpu_ack = H;

`ifndef N6502
//****************************************************************************
output cpu_hiram, cpu_loram, cpu_charen;

//dma cannot write this register, it can write to the ram image though
assign cpu_w = ~(|cpu_a[15:1]) & cpu_a[0] & ~cpu_rw & ~cpu_ack;

//C64 cpu memory control register
reg [2:0] a1 = 3'h6;
assign cpu_loram = a1[0];
assign cpu_hiram = a1[1];
assign cpu_charen = a1[2];

always@(posedge clk)
  begin
    //update the cpu memory map control register
    if(~resetIn)   a1 <= 3'h6; //set it wrong just to know it changes
    else if(cpu_w) a1 <= cpu_wb[2:0];
  end
`endif             

//****************************************************************************
// parameter definitions
parameter V_NMI = 16'hFFFA;
parameter V_RST = 16'hFFFC;
parameter V_IRQ = 16'hFFFE;
parameter ZERO = 8'h00;
parameter ONE = 8'h01;
parameter NEG_ONE = 8'hFF;
parameter H = 1'b1;
parameter L = 1'b0;
parameter RD = 1'b1;
parameter WR = 1'b0;
parameter TRUE = 1'b1;
parameter FALSE = 1'b0;
parameter BYTEX = 4'hF;
parameter BYTE0 = 4'hE;
parameter BYTE1 = 4'hD;
parameter BYTE2 = 4'hB;
parameter BYTE3 = 4'h7;
parameter BYTE01 = 4'hC;
parameter BYTE12 = 4'h9;
parameter BYTE23 = 4'h3;
parameter BYTE30 = 4'h6;
parameter BYTE012 = 4'h8;
parameter BYTE123 = 4'h1;
parameter BYTE230 = 4'h2;
parameter BYTE301 = 4'h4;

//size of instruction queue
parameter IR_QUEUE_SIZE = 4'h9;

//****************************************************************************
//cpu state machine state names
//
// put reset vector on address bus to read the startup pc
parameter S_RESET = 4'h0;
// load IR with first instruction bytes
parameter S_LOAD_INST = 4'h1;
// decode the first instruction of new code path
parameter S_DECODE_INST = 4'h2;
// execute 1 byte instruction
parameter S_EXECUTE_INST1 = 4'h3;
// execute 2 byte instruction
parameter S_EXECUTE_INST2 = 4'h4;
// execute 3 byte instruction
parameter S_EXECUTE_INST3 = 4'h5;
// indirect y operand base address load
parameter S_IND_Y = 4'h6;
// read memory operand - indirect addressing modes only
parameter S_RD_MEM_OP = 4'h7;
// write memory operand, update flags when necessary
parameter S_WR_MEM_OP = 4'h8;
// execute memory to register operation, update flags
parameter S_REG_OP = 4'h9;
// give BCD arithmatic an extra cycle to complete
parameter S_BCD_DELAY = 4'hA;
// rts pc adjust (+1)
parameter S_RTS = 4'hB;
// process nmi or irq
parameter S_INT = 4'hC;
// self modifying code wrote to a byte in the next instruction
parameter S_IR_HIT = 4'hD;
// load more instruction bytes
parameter S_IR_LOAD = 4'hE;
// delayed rmw 3 byte instruction
parameter S_RD_MEM_OP2 = 4'hF;

//state machine
reg [3:0] next_state, state = S_RESET;

//****************************************************************************
// State machine patterns...
//
// [***] Instruction sequences below end in going to one of the S_EXECUTE_INSTn
// states. S_IR_LOAD or S_IR_HIT may occur before entering S_EXECUTE_INSTn under
// certain conditions.
//
// [S_INT] indicates starting execution of either the irq or the nmi exception
// state sequences and replaces the normal subsequent states listed.
// The brk & rti instructions do not respond to the irq or nmi signals.
//
// [S_IR_LOAD] indicates a conditional extra state is required to
// keep the instruction decoding process going. The next instruction to be
// executed must have a decode cycle done with enough bytes for the entire
// instruction available in the instruction register before one of the
// S_EXECUTE_INSTn states can be entered.
//
// [S_IR_HIT] indicates self modifying code has just written to the next
// instruction in the ir register so an extra cycle is required for decoding
// the new information.
//
// default cpu_a  address XXX == {cpu_if, L} - next 4 bytes in instruction stream
// default cpu_sa address XXX == {ONE, spp1} - current value on the stack
//
// {cpu_if, L} accesses will only use the data and increment itself if
// there is room in the ir queue for 4 more bytes.
//
// # indicates a write cycle
//

//****************************************************************************
// EXCEPTIONAL EVENTS :
//
// reset : ? / 3
// state  : S_RESET -> S_LOAD_INST -> S_DECODE_INST -> [***]
// cpu_a  : <V_RST>    <pc>           {cpu_if, L}
// cpu_ia : XXX        XXX            XXX

// irq : sp gets decremented by 3 in the first state 7+ / 3+ ?
// state  : S_INT ->     S_LOAD_INST -> S_DECODE_INST -> [***]
// cpu_a  : <V_IRQ>      <pc>           {cpu_if, L}
// cpu_ia : {ONE, spm2}# XXX            XXX

// nmi : sp gets decremented by 3 in the first state 7+ / 3+ ?
// state  : S_INT ->     S_LOAD_INST -> S_DECODE_INST -> [***]
// cpu_a  : <V_NMI>      <pc>           {cpu_if, L}
// cpu_ia : {ONE, spm2}# XXX            XXX

// ir re-decode due to self modifying code : - / 1
// state  : S_IR_HIT    -> [S_IR_LOAD] [***]
// cpu_a  : {cpu_if, L}
// cpu_ia : XXX

// ir load due to insufficient queue bytes : - / 1+
// state  : S_IR_LOAD   -> [S_IR_LOAD] [***]
// cpu_a  : {cpu_if, L}
// cpu_ia : XXX

//****************************************************************************
// ONE BYTE INSTRUCTIONS :
//
// brk : 7 / 3
// state  : S_EXECUTE_INST1 -> [S_INT] [S_LOAD_INST -> S_DECODE_INST -> ***]
// cpu_a  : <V_IRQ>                    <pc>           {cpu_if, L}
// cpu_ia : {ONE, spm2}#               XXX            XXX

// rti : 6 / 3
// state  : S_EXECUTE_INST1 -> [S_INT] [S_LOAD_INST -> S_DECODE_INST -> ***]
// cpu_a  : XXX                        <pc>           {cpu_if, L}
// cpu_ia : {ONE, spp1}                XXX            XXX

// rts : 6 / 4 - in S_RTS the pc off the stack is incremented
// state  : S_EXECUTE_INST1 -> S_RTS -> [S_INT] [S_LOAD_INST -> S_DECODE_INST -> ***]
// cpu_a  : XXX                XXX               <pc>           {cpu_if, L}
// cpu_ia : {ONE, spp1}        XXX               XXX            XXX

// php, pha : 3 / 1
// state  : S_EXECUTE_INST1 -> [S_INT] [***]
// cpu_a  : {cpu_if, L}
// cpu_ia : {ONE, sp}#

// pla : 4 / 2 - S_REG_OP is where flags are updated
// state  : S_EXECUTE_INST1 -> S_REG_OP -> [S_INT] [***]
// cpu_a  : {cpu_if, L}        {cpu_if, L}
// cpu_ia : {ONE, spp1}        XXX

// plp : 4 / 1
// state  : S_EXECUTE_INST1 -> [S_INT] [***]
// cpu_a  : {cpu_if, L}
// cpu_ia : {ONE, spp1}

// all other one byte instructions : 2 / 1
// state  : S_EXECUTE_INST1 -> [S_INT] [***]
// cpu_a  : {cpu_if, L}
// cpu_ia : XXX

//****************************************************************************
// TWO BYTE INSTRUCTIONS :
//
// branch_taken : 3+ / 2
// state  : S_EXECUTE_INST2 -> [S_INT] [S_DECODE_INST -> ***]
// cpu_a  : <branch>                    {cpu_if, L}
// cpu_ia : XXX                         XXX

// (sta_ix) 6 / 2
// state  : S_EXECUTE_INST2 -> S_WR_MEM_OP -> [S_INT] [S_IR_HIT] [S_IR_LOAD] [***]
// cpu_a  : {cpu_if, L}        <tmp_a>#        
// cpu_ia : {ZERO,addx[7:0]}   <tmp_a>#      

// (_ix) 6 / 4 -  adc,sbc in decimal mode
// state  : S_EXECUTE_INST2 -> S_RD_MEM_OP -> S_REG_OP -> S_BCD_DELAY -> [S_INT] [***]
// cpu_a  : {cpu_if, L}        <tmp_a>        {cpu_if, L} {cpu_if, L}
// cpu_ia : {ZERO,addx[7:0]}   <tmp_a>        XXX         XXX

// (_ix) 6 / 3 -  all other register operations
// state  : S_EXECUTE_INST2 -> S_RD_MEM_OP -> S_REG_OP -> [S_INT] [***]
// cpu_a  : {cpu_if, L}        <tmp_a>        {cpu_if, L} 
// cpu_ia : {ZERO,addx[7:0]}   <tmp_a>        XXX

// (sta_iy) 6 / 3 - S_IND_Y used to add the y index
// state  : S_EXECUTE_INST2 -> S_IND_Y ->  S_WR_MEM_OP -> [S_INT] [S_IR_HIT] [S_IR_LOAD] [***]
// cpu_a  : {cpu_if, L}        {cpu_if, L} <tmp_a>#
// cpu_ia : {ZERO,add[7:0]}    XXX         <tmp_a>#

// (_iy) 5+ / 5 - adc,sbc in decimal mode
// state  : S_EXECUTE_INST2 -> S_IND_Y ->  S_RD_MEM_OP -> S_REG_OP -> S_BCD_DELAY -> [S_INT] [***]
// cpu_a  : {cpu_if, L}        {cpu_if, L} <tmp_a>        {cpu_if, L} {cpu_if, L}
// cpu_ia : {ZERO,add[7:0]}    XXX         <tmp_a>        XXX         XXX

// (_iy) 5+ / 4 - all other register operations
// state  : S_EXECUTE_INST2 -> S_IND_Y ->  S_RD_MEM_OP -> S_REG_OP -> [S_INT] [***]
// cpu_a  : {cpu_if, L}        {cpu_if, L} <tmp_a>        {cpu_if, L} 
// cpu_ia : {ZERO,add[7:0]}    XXX         <tmp_a>        XXX

// sta_z,stx_z,sty_z : 3 / 2 (tmp_a ==  {ZERO,add[7:0]})
// state  : S_EXECUTE_INST2 -> S_WR_MEM_OP -> [S_INT] [S_IR_HIT] [S_IR_LOAD] [***]
// cpu_a  : {cpu_if, L}        <tmp_a>#
// cpu_ia : XXX                <tmp_a>#

// _z : 5 / 2 - memory operations
// state  : S_EXECUTE_INST2 -> S_WR_MEM_OP -> [S_INT] [S_IR_HIT] [S_IR_LOAD] [***]
// cpu_a  : {cpu_if, L}        {ZERO,add[7:0]}#
// cpu_ia : {ZERO,add[7:0]}    {ZERO,add[7:0]}#

// _z : 3 / 3 - adc,sbc in decimal mode
// state  : S_EXECUTE_INST2 -> S_REG_OP -> S_BCD_DELAY -> [S_INT] [***]
// cpu_a  : {cpu_if, L}        {cpu_if, L} {cpu_if, L}
// cpu_ia : {ZERO,add[7:0]}    XXX         XXX

// _z : 3 / 2 - all other register operations
// state  : S_EXECUTE_INST2 -> S_REG_OP -> [S_INT] [***]
// cpu_a  : {cpu_if, L}        {cpu_if, L}
// cpu_ia : {ZERO,add[7:0]}    XXX

// sta_zx,sty_zx : 4 / 2 - (tmp_a == {ZERO,addx[7:0]})
// state  : S_EXECUTE_INST2 -> S_WR_MEM_OP -> [S_INT] [S_IR_HIT] [S_IR_LOAD] [***]
// cpu_a  : {cpu_if, L}        <tmp_a>#
// cpu_ia : XXX                <tmp_a>#

// _zx : 6 / 2 - memory operations
// state  : S_EXECUTE_INST2 -> S_WR_MEM_OP -> [S_INT] [S_IR_HIT] [S_IR_LOAD] [***]
// cpu_a  : {cpu_if, L}        {ZERO,addx[7:0]}#
// cpu_ia : {ZERO,addx[7:0]}   {ZERO,addx[7:0]}#

// _zx : 4 / 3 - adc,sbc in decimal mode
// state  : S_EXECUTE_INST2 -> S_REG_OP -> S_BCD_DELAY -> [S_INT] [***]
// cpu_a  : {cpu_if, L}        {cpu_if, L} {cpu_if, L}
// cpu_ia : {ZERO,addx[7:0]}   XXX         XXX

// _zx : 4 / 2 - all other register operations
// state  : S_EXECUTE_INST2 -> S_REG_OP -> [S_INT] [***]
// cpu_a  : {cpu_if, L}        {cpu_if, L}
// cpu_ia : {ZERO,addx[7:0]}   XXX

// stx_zy : 4 / 2 - (tmp_a == {ZERO,addy[7:0]})
// state  : S_EXECUTE_INST2 -> S_WR_MEM_OP -> [S_INT] [S_IR_HIT] [S_IR_LOAD] [***]
// cpu_a  : {cpu_if, L}        <tmp_a>#
// cpu_ia : XXX                <tmp_a>#

// ldx_zy : 4 / 2 - register load
// state  : S_EXECUTE_INST2 -> S_REG_OP -> [S_INT] [***]
// cpu_a  : {cpu_if, L}        {cpu_if, L}
// cpu_ia : {ZERO,addy[7:0]}   XXX

// adc,sbc : 2 / 2 - immediate instructions in decimal mode
// state  : S_EXECUTE_INST2 -> S_BCD_DELAY -> [S_INT] [***]
// cpu_a  : {cpu_if, L}        {cpu_if, L}
// cpu_ia : XXX                XXX

// all other immediate instructions and branches not taken : 2 / 1
// state  : S_EXECUTE_INST2 -> [S_INT] [***]
// cpu_a  : {cpu_if, L}
// cpu_ia : XXX

//****************************************************************************
// THREE BYTE INSTRUCTIONS :
//
// jmp_i : 5 / 3
// state  : S_EXECUTE_INST3 -> [S_INT] [S_LOAD_INST -> S_DECODE_INST -> ***]
// cpu_a  : <add>                       <pc>           {cpu_if, L}
// cpu_ia : <add>                       XXX            XXX

// jmp_a : 3 / 2
// state  : S_EXECUTE_INST3 -> [S_INT] [S_DECODE_INST -> ***]
// cpu_a  : <add>                       {cpu_if, L}
// cpu_ia : XXX                         XXX

// jsr_a : 6 / 2
// state  : S_EXECUTE_INST3 -> [S_INT] [S_DECODE_INST -> ***]
// cpu_a  : <add>                       {cpu_if, L}
// cpu_ia : {ONE, spm2}#                XXX

// sta_a,stx_a,sty_a : 4 / 2 (tmp_a == add)
// state  : S_EXECUTE_INST3 -> S_WR_MEM_OP -> [S_INT] [S_IR_HIT] [S_IR_LOAD] [***]
// cpu_a  : <{cpu_if, L}       <tmp_a>#
// cpu_ia : XXX                <tmp_a>#

// _a : 6 / 3 - memory operations (tmp_a == add)
// state  : S_EXECUTE_INST3 -> S_RD_MEM_OP2 -> S_WR_MEM_OP -> [S_INT] [S_IR_HIT] [S_IR_LOAD] [***]
// cpu_a  : {cpu_if, L}        <tmp_a>         <tmp_a>#
// cpu_ia : XXX                <tmp_a>         <tmp_a>#

// _a : 4 / 3 - adc,sbc in decimal mode
// state  : S_EXECUTE_INST3 -> S_REG_OP  -> S_BCD_DELAY -> [S_INT] [***]
// cpu_a  : <add>              {cpu_if, L}  {cpu_if, L}
// cpu_ia : <add>              XXX           XXX

// _a : 4 / 2 - all other register operations
// state  : S_EXECUTE_INST3 -> S_REG_OP -> [S_INT] [***]
// cpu_a  : <add>             {cpu_if, L}
// cpu_ia : <add>             XXX

// sta_ax : 5 / 2 (tmp_a == addx)
// state  : S_EXECUTE_INST3 -> S_WR_MEM_OP -> [S_INT] [S_IR_HIT] [S_IR_LOAD] [***]
// cpu_a  : {cpu_if, L}        <tmp_a>#
// cpu_ia : XXX                <tmp_a>#

// _ax : 7 / 3 - memory operations (tmp_a == addx)
// state  : S_EXECUTE_INST3 -> S_RD_MEM_OP2 -> S_WR_MEM_OP -> [S_INT] [S_IR_LOAD] [***]
// cpu_a  :{cpu_if, L}         <tmp_a>         <tmp_a>#
// cpu_ia : XXX                <tmp_a>         <tmp_a>#

// _ax : 4+ / 3 - adc,sbc in decimal mode
// state  : S_EXECUTE_INST3 -> S_REG_OP  -> S_BCD_DELAY -> [S_INT] [***]
// cpu_a  : <addx>             {cpu_if, L}  {cpu_if, L}
// cpu_ia : <addx>             XXX          XXX

// _ax : 4+ / 2 - all other register operations
// state  : S_EXECUTE_INST3 -> S_REG_OP -> [S_INT] [***]
// cpu_a  : <addx>             {cpu_if, L}
// cpu_ia : <addx>             XXX

// sta_ay : 5 / 2 (tmp_a == addy)
// state  : S_EXECUTE_INST3 -> S_WR_MEM_OP -> [S_INT] [S_IR_HIT] [S_IR_LOAD] [***]
// cpu_a  : {cpu_if, L}        <tmp_a>#
// cpu_ia : XXX                <tmp_a>#

// _ay : 4+ / 3 - adc,sbc in decimal mode
// state  : S_EXECUTE_INST3 -> S_REG_OP -> S_BCD_DELAY -> [S_INT] [***]
// cpu_a  : <addy>             {cpu_if, L} {cpu_if, L}
// cpu_ia : <addy>             XXX          XXX

// _ay : 4+ / 2 - all other register operations
// state  : S_EXECUTE_INST3 -> S_REG_OP -> [S_INT] [***]
// cpu_a  : <addy>             {cpu_if, L}
// cpu_ia : <addy>             XXX

//****************************************************************************
//status register bits
reg negative, zero, carry, overflow, interrupt, decimal, next_N, next_Z, next_C, next_V;

//next instruction registers
reg [7:0] ira, imm_op, ah;

//two_byte_a - this may not indicate a valid 2 byte instruction, it does
//indicate it is not a valid one or three byte instruction...
//two_byte - this indicates a valid two byte instruction after a decoding cycle

iDecode id(clk, RDY, ira, next_C, next_V, next_N, next_Z,
           two_byte, one_byte_a, two_byte_a, three_byte_a, branch_taken,
           asla, rola, lsra, rora, dex, dey, inx, iny,txa, tya, txs, tay,
           tax, tsx, pha, pla, clc, sec, cli, sei, cld, sed, clv, php, plp,
           ora_i, anda_i, eora_i, lda_i, cmpa_i,// adca_i, sbca_i,
           ora_z, anda_z, eora_z, lda_z, cmpa_z, bita_z,// adca_z, sbca_z,
           ora_zx, anda_zx, eora_zx, lda_zx, cmpa_zx,// adca_zx, sbca_zx,
           sta_ix, ora_ix, anda_ix, eora_ix, lda_ix, cmpa_ix,// adca_ix, sbca_ix,
           sta_iy, ora_iy, anda_iy, eora_iy, lda_iy, cmpa_iy,// adca_iy, sbca_iy,
           ora_a, anda_a, eora_a, lda_a, cmpa_a, bita_a,// adca_a, sbca_a,
           ora_ax, anda_ax, eora_ax, ldy_ax, lda_ax, cmpa_ax,// adca_ax, sbca_ax,
           ora_ay, anda_ay, eora_ay, lda_ay, ldx_ay, cmpa_ay,// adca_ay, sbca_ay,
           asl_z, rol_z, lsr_z, ror_z, dec_z, inc_z,
           asl_zx, rol_zx, lsr_zx, ror_zx, dec_zx, inc_zx,
           asl_a, rol_a, lsr_a, ror_a, dec_a, inc_a,
           asl_ax, rol_ax, lsr_ax, ror_ax, dec_ax, inc_ax,
           stx_z, ldx_z, cpx_z, stx_zy, ldx_zy, stx_a, ldx_a, cpx_a, ldx_i, cpx_i,
           sty_z, ldy_z, cpy_z, sty_zx, ldy_zx, sty_a, ldy_a, cpy_a, ldy_i, cpy_i,
           brk, rti, rts, jmp_a, jmp_i, jsr_a,
           //sta_z, sta_zx, //sta_a, sta_ax, sta_ay,
           ix, iy, st_zxy, st_axy, zx, zy, ax, ay, mem_op, reg_op, addsub, imm_addsub);

//decoded instructions saved for use out of the S_EXECUTE_INSTn states
reg sta_iy_2, stx_2, sty_2, addsub_2;
reg asl, rol, ror, lsr, inc, dec;
reg ora, bita, anda, eora, ldx, lda, ldy, cpx_2, cmpa_2, cpy_2;

//data and index registers
reg [7:0] a, x, y, alu_op, next_x, next_y;

//stack pointer addresses
reg [7:0] sp, spp1, spp2, spp3, next_sp;
wire [7:0] next_spp1, next_spp2, next_spp3, spm1, spm2, spm3;
assign next_spp1 = next_sp + H;
assign next_spp2 = next_sp + 2'h2;
assign next_spp3 = next_sp + 2'h3;
assign spm1 = sp - H;
assign spm2 = sp - 2'h2;
assign spm3 = sp - 2'h3;

//processor status stacking options
wire [7:0] sp_int, sp_brk;

//                N         V         X  B  D        I          Z     C
assign sp_int = { negative, overflow, H, L, decimal, interrupt, zero, carry };
assign sp_brk = { negative, overflow, H, H, decimal, interrupt, zero, carry };

//instruction registers
reg [15:0] branch, next_pc, pc, pc_p1r, pc_p2r, pc_p3r, pc_p4r, pc_p5r, pc_p6r, pc_p7r, pc_p8r;

//program counter increment options
wire [15:0] pcp1, pcp2, pcp3, pcp4, pcp5, pcp6, pcp7, pcp8;//, pcp9, pcpA;
assign pcp1 = next_pc + 16'h0001;
assign pcp2 = next_pc + 16'h0002;
assign pcp3 = next_pc + 16'h0003;
assign pcp4 = next_pc + 16'h0004;
assign pcp5 = next_pc + 16'h0005;
assign pcp6 = next_pc + 16'h0006;
assign pcp7 = next_pc + 16'h0007;
assign pcp8 = next_pc + 16'h0008;

wire [15:0] branchp0, branchp1, branchp2;
assign branchp0 = pc_p2r + { {9{imm_op[7]}}, imm_op[6:0] };
assign branchp1 = pc_p3r + { {9{imm_op[7]}}, imm_op[6:0] };
assign branchp2 = pc_p4r + { {9{imm_op[7]}}, imm_op[6:0] };

//instruction fetch address register and control;
//bit 0 never used...accesses always on word boundaries (for 4 usable bytes)
reg [15:1] cpu_if, next_if;
wire [15:1] cpu_ifp4 = cpu_if + 15'h0002;

//11,12 or 15 bytes of instruction queue register
reg [IR_QUEUE_SIZE*8-1:0] ir;       
reg [3:0] iCount;     //how many instruction bytes currently in the ir register

//for making sure we can decode the instruction we think that we have (or more implied)
reg one_valid, two_valid, three_valid, four_valid, five_valid;

//instruction data buses
wire [7:0] ib0, ib1, ib2, ib3;
wire [31:0] ib32;
assign ib0 = cpu_if[1] ? cpu_rd[23:16] : cpu_rd[7:0];
assign ib1 = cpu_if[1] ? cpu_rd[31:24] : cpu_rd[15:8];
assign ib2 = cpu_if[1] ? cpu_rd[7:0]   : cpu_rd[23:16];
assign ib3 = cpu_if[1] ? cpu_rd[15:8]  : cpu_rd[31:24];
assign ib32 = { ib3, ib2, ib1, ib0 };

//if low an add is occurring else it is a sub or cmp
reg alu_mode;

//alu outputs
wire [7:0] result, asl_op, rol_op, lsr_op, ror_op, dec_op, inc_op, or_op, and_op, eor_op, bcdResult;

//binary alu
accumulatorPlus aA(alu_mode, a, alu_op, carry, result, Cout_a, is_zero_a, is_negative_a, is_overflow_a,
                   alu_op_0, alu_op_1, alu_op_255, alu_op_dec_neg, alu_op_inc_neg,
                   asl_op, rol_op, lsr_op, ror_op, dec_op, inc_op,
                   or_op, and_op, eor_op, or_op_zero, and_op_zero, eor_op_zero);
                   
//decimal mode alu
bcdAccumulator aBCD(alu_mode, a, alu_op, carry, bcdResult, bcdCout_a, bcdZero_a, bcdNegative_a, bcdOverflow_a);

//compare alu
compare cA(a, alu_op, Cout_ca, is_zero_ca, N_ca);
compare cX(x, alu_op, Cout_x, is_zero_x, N_x);
compare cY(y, alu_op, Cout_y, is_zero_y, N_y);

//write byte data register
reg [31:0] cpu_wsb;
reg [7:0] cpu_wb;
reg cpu_rw, cpu_rwz, cpu_rws;

//byte select write control register
reg [3:0] cpu_wbyte, cpu_wbytez, cpu_wbytes;

//operand addresses
reg [15:0] addx, addy, tmp_a;

wire [15:0] add_x = { ah, imm_op } + next_x;
wire [15:0] add_y = { ah, imm_op } + next_y;
wire [15:0] add = ir[23:8];

//memory bus addreseses
reg [15:0] cpu_a, cpu_ia;

//for the second address bus for lower 16 bit external data bus
reg [15:2] cpu_al, addl, addxl, addyl, ifl, branchl, pc_rl;

//lower 16 bit internal data bus address generation 
wire [7:2] cpu_ial = cpu_ia[7:2] + cpu_ia[1];

//stack / zero page access hit
assign internal_hit = ~(|cpu_ia[15:9]);

wire [31:0] internal, internal_s, internal_z;

ram256 zerop(clk, cpu_wbytez, ~cpu_ia[8] & internal_hit, cpu_ia[7:2], cpu_ial, { cpu_wb, cpu_wb, cpu_wb, cpu_wb }, internal_z);
ram256 stack(clk, cpu_wbytes,  cpu_ia[8] & internal_hit, cpu_ia[7:2], cpu_ial, cpu_wsb, internal_s);

//combined zp and stack memory data
assign internal = cpu_ia[8] ? internal_s : internal_z;

//address read from the stack
wire [15:0] cpu_sar0, cpu_sar1, cpu_sar2, cpu_sar3;
assign cpu_sar0 = internal_s[15:0];
assign cpu_sar1 = internal_s[23:8];
assign cpu_sar2 = internal_s[31:16];
assign cpu_sar3 = { internal_s[7:0], internal_s[31:24] };

//address read from zero page
wire [15:0] cpu_zar0, cpu_zar1, cpu_zar2, cpu_zar3;
assign cpu_zar0 = internal_z[15:0];
assign cpu_zar1 = internal_z[23:8];
assign cpu_zar2 = internal_z[31:16];
assign cpu_zar3 = { internal_z[7:0], internal_z[31:24] };

//address read from external memory
wire [15:0] cpu_dar0, cpu_dar1, cpu_dar2, cpu_dar3;
assign cpu_dar0 = cpu_rd[15:0];
assign cpu_dar1 = cpu_rd[23:8];
assign cpu_dar2 = cpu_rd[31:16];
assign cpu_dar3 = { cpu_rd[7:0], cpu_rd[31:24] };

//synchronizing interrupt registers
reg rirq = H;       //synchonize irq
reg rnmi = H;       //synchonize nmi
reg rnmi2 = H;      //this is used to determine a falling rnmi edge
reg nmi = H;        //this is used to save the falling edge occurance until the next instruction

//an unimplemented instruction or some other fatal event causes at least
//1 bit in this variable to be set low and stops the processor in a halt condition
reg [3:0] halt = 4'hF;

//asynchronous control state machines
always@(*)
  begin
//****************************************************************************
//****************************************************************************
//ir decode
    if(~cpu_ack)
      begin
        case(state)
//****************************************************************************
          S_EXECUTE_INST1:
            begin
              ira = ir[15:8];
              imm_op = ir[23:16];
              ah = ir[31:24];
            end
//****************************************************************************
          S_EXECUTE_INST2:
            begin
              ira = ir[23:16];
              imm_op = ir[31:24];
              ah = ir[39:32];
            end
//****************************************************************************
//          S_EXECUTE_INST3:
//            begin
//              ira = ir[31:24];
//              imm_op = ir[39:32];
//              ah = ir[47:40];
//            end
//****************************************************************************
          default:
            begin
              ira = ir[7:0];
              imm_op = ir[15:8];
              ah = ir[23:16];
            end
        endcase
      end   
    else
      begin
        ira = ir[7:0];
        imm_op = ir[15:8];
        ah = ir[23:16];
      end

//****************************************************************************
//****************************************************************************
//write bus
    if(~cpu_ack)
      begin
        case(state)
//****************************************************************************
          S_EXECUTE_INST2:
            begin
              if(stx_z|stx_zy)      cpu_wb = x;
              else if(sty_z|sty_zx) cpu_wb = y;
              else                  cpu_wb = a;
            end
//****************************************************************************
          S_WR_MEM_OP: // complete memory operation
            begin
              if(asl)               cpu_wb = asl_op;
              else if(rol)          cpu_wb = rol_op;
              else if(lsr)          cpu_wb = lsr_op;
              else if(ror)          cpu_wb = ror_op;
              else if(dec)          cpu_wb = dec_op;
              else if(inc)          cpu_wb = inc_op;
              else if(stx_2)        cpu_wb = x;
              else if(sty_2)        cpu_wb = y;
              else                  cpu_wb = a;
            end
//****************************************************************************
          default:                  cpu_wb = a;
        endcase
      end
    else                            cpu_wb = dma_dw;

//****************************************************************************
//****************************************************************************
//stack write bus
    if(~cpu_ack)
      begin
        case(state)
//****************************************************************************
          S_INT:
            case(spm2[1:0])
              0 :             cpu_wsb = { a, pc[15:8], pc[7:0], sp_int };
              1 :             cpu_wsb = { pc[15:8], pc[7:0], sp_int, a };
              2 :             cpu_wsb = { pc[7:0], sp_int, a, pc[15:8] };
              3 :             cpu_wsb = { sp_int, a, pc[15:8], pc[7:0] };
            endcase
//****************************************************************************
          S_EXECUTE_INST1:
            begin
              if(brk)
                case(spm2[1:0])
                  0 :         cpu_wsb = { a, pc[15:8], pc[7:0], sp_brk };
                  1 :         cpu_wsb = { pc[15:8], pc[7:0], sp_brk, a };
                  2 :         cpu_wsb = { pc[7:0], sp_brk, a, pc[15:8] };
                  3 :         cpu_wsb = { sp_brk, a, pc[15:8], pc[7:0] };
                endcase
              else if(php)    cpu_wsb = { sp_brk, sp_brk, sp_brk, sp_brk };
              else            cpu_wsb = { a, a, a, a };
            end
//****************************************************************************
          S_EXECUTE_INST3:
            begin
              if(jsr_a)
                case(spm2[1:0])
                  0 :         cpu_wsb = { a, pc_p2r[15:8], pc_p2r[7:0], a };
                  1 :         cpu_wsb = { pc_p2r[15:8], pc_p2r[7:0], a, a };
                  2 :         cpu_wsb = { pc_p2r[7:0], a, a, pc_p2r[15:8] };
                  3 :         cpu_wsb = { a, a, pc_p2r[15:8], pc_p2r[7:0] };
                endcase
              else            cpu_wsb = { a, a, a, a };
            end
//****************************************************************************
          S_WR_MEM_OP: // complete memory operation
            begin
              if(asl)         cpu_wsb = { asl_op, asl_op, asl_op, asl_op };
              else if(rol)    cpu_wsb = { rol_op, rol_op, rol_op, rol_op };
              else if(lsr)    cpu_wsb = { lsr_op, lsr_op, lsr_op, lsr_op };
              else if(ror)    cpu_wsb = { ror_op, ror_op, ror_op, ror_op };
              else if(dec)    cpu_wsb = { dec_op, dec_op, dec_op, dec_op };
              else if(inc)    cpu_wsb = { inc_op, inc_op, inc_op, inc_op };
              else if(stx_2)  cpu_wsb = { x, x, x, x };
              else if(sty_2)  cpu_wsb = { y, y, y, y };
              else            cpu_wsb = { a, a, a, a };
            end
//****************************************************************************
          default:            cpu_wsb = { a, a, a, a };
        endcase
      end
    else                      cpu_wsb = { dma_dw, dma_dw, dma_dw, dma_dw };

//****************************************************************************
//****************************************************************************
// write or read instruction operand address bus
    if(~cpu_ack)
      begin
        case(state)
//****************************************************************************
          S_EXECUTE_INST1:
            begin
              if(brk)                           cpu_ia = { ONE, spm2 };
              else if(pha|php)                  cpu_ia = { ONE, sp };
              else                              cpu_ia = { ONE, spp1 };
            end
//****************************************************************************
          S_EXECUTE_INST2:
            begin
              if(zx|ix)                         cpu_ia = { ZERO, addx[7:0] };
              else if(zy)                       cpu_ia = { ZERO, addy[7:0] };
              else                              cpu_ia = { ZERO, add[7:0] };
            end
//****************************************************************************
          S_EXECUTE_INST3:
            begin
              if(jsr_a)                         cpu_ia = { ONE, spm2 };
              else if(ax)                       cpu_ia = addx;
              else if(ay)                       cpu_ia = addy;
              else                              cpu_ia = add;
            end
//****************************************************************************
          S_RD_MEM_OP,S_RD_MEM_OP2,S_WR_MEM_OP: cpu_ia = tmp_a;
//****************************************************************************
          S_INT:                                cpu_ia = { ONE, spm2 };
//****************************************************************************
          default:                              cpu_ia = { ONE, spp1 };
        endcase
      end
    else                                        cpu_ia = dma_a;

//****************************************************************************
//****************************************************************************
//separate r/w control of the internal memory
    if(~cpu_ack)
      begin
        case(state)
//****************************************************************************
          S_WR_MEM_OP: cpu_rwz = WR;
//****************************************************************************
          default:     cpu_rwz = RD;
        endcase
      end
    else if(~dma_ack)  cpu_rwz = dma_rw;
    else               cpu_rwz = RD;

//****************************************************************************
//****************************************************************************
//write byte select
    if(~cpu_rwz)
      begin
        case(cpu_ia[1:0])
          2'h0 : cpu_wbytez = BYTE0;
          2'h1 : cpu_wbytez = BYTE1;
          2'h2 : cpu_wbytez = BYTE2;
          2'h3 : cpu_wbytez = BYTE3;
        endcase
      end
    else         cpu_wbytez = BYTEX;

//****************************************************************************
//****************************************************************************
//separate r/w control of the stack memory
    if(~cpu_ack)
      begin
        case(state)
//****************************************************************************
          S_EXECUTE_INST1:
            begin
              if(brk|php|pha) cpu_rws = WR;
              else            cpu_rws = RD;
            end
//****************************************************************************
          S_EXECUTE_INST3:
            begin
              if(jsr_a)       cpu_rws = WR;
              else            cpu_rws = RD;
            end
//****************************************************************************
          S_WR_MEM_OP:        cpu_rws = WR; // only if in stack page
//****************************************************************************
          S_INT:              cpu_rws = WR;
//****************************************************************************
          default:            cpu_rws = RD;
        endcase
      end
    else if(~dma_ack)         cpu_rws = dma_rw;
    else                      cpu_rws = RD;

//****************************************************************************
//****************************************************************************
//stack write bus
    if(~cpu_ack)
      begin
        case(state)
//****************************************************************************
          S_INT:
            case(spm2[1:0])
              0 :     cpu_wbytes = BYTE012;
              1 :     cpu_wbytes = BYTE123;
              2 :     cpu_wbytes = BYTE230;
              3 :     cpu_wbytes = BYTE301;
            endcase
//****************************************************************************
          S_EXECUTE_INST1:
            begin
              if(brk)
                case(spm2[1:0])
                  0 : cpu_wbytes = BYTE012;
                  1 : cpu_wbytes = BYTE123;
                  2 : cpu_wbytes = BYTE230;
                  3 : cpu_wbytes = BYTE301;
                endcase
              else if(~cpu_rws)
                case(cpu_ia[1:0])
                  0 : cpu_wbytes = BYTE0;
                  1 : cpu_wbytes = BYTE1;
                  2 : cpu_wbytes = BYTE2;
                  3 : cpu_wbytes = BYTE3;
                endcase
              else    cpu_wbytes = BYTEX;
            end
//****************************************************************************
          S_EXECUTE_INST3:
            begin
              if(jsr_a)
                case(spm2[1:0])
                  0 : cpu_wbytes = BYTE12;
                  1 : cpu_wbytes = BYTE23;
                  2 : cpu_wbytes = BYTE30;
                  3 : cpu_wbytes = BYTE01;
                endcase
              else if(~cpu_rws)
                case(cpu_ia[1:0])
                  0 : cpu_wbytes = BYTE0;
                  1 : cpu_wbytes = BYTE1;
                  2 : cpu_wbytes = BYTE2;
                  3 : cpu_wbytes = BYTE3;
                endcase
              else    cpu_wbytes = BYTEX;
            end
//****************************************************************************
          default://      cpu_wbytes = BYTEX;
              if(~cpu_rws)
                begin
                  case(cpu_ia[1:0])
                    0 : cpu_wbytes = BYTE0;
                    1 : cpu_wbytes = BYTE1;
                    2 : cpu_wbytes = BYTE2;
                    3 : cpu_wbytes = BYTE3;
                  endcase
                end
              else      cpu_wbytes = BYTEX;
        endcase
      end
    else if(~dma_ack & ~dma_rw)
      begin
        case(cpu_ia[1:0])
          0 :           cpu_wbytes = BYTE0;
          1 :           cpu_wbytes = BYTE1;
          2 :           cpu_wbytes = BYTE2;
          3 :           cpu_wbytes = BYTE3;
        endcase
      end
    else                cpu_wbytes = BYTEX;


//****************************************************************************
//****************************************************************************
//r/w control (of the write bus)
    if(~cpu_ack)
      begin
        case(state)
//****************************************************************************
          S_WR_MEM_OP: cpu_rw = WR;
//****************************************************************************
          default:     cpu_rw = RD;
        endcase
      end
    else if(~dma_ack)  cpu_rw = dma_rw;
    else               cpu_rw = RD;

//****************************************************************************
//****************************************************************************
//write byte select
    if(~cpu_rw)
      begin
        case(cpu_a[1:0])
          2'h0 : cpu_wbyte = BYTE0;
          2'h1 : cpu_wbyte = BYTE1;
          2'h2 : cpu_wbyte = BYTE2;
          2'h3 : cpu_wbyte = BYTE3;
        endcase
      end
    else         cpu_wbyte = BYTEX;

//****************************************************************************
//****************************************************************************
//read instruction, pc vector address bus
    if(~cpu_ack)
      begin
        case(state)
//****************************************************************************
          S_RESET:                              cpu_a = V_RST;
//****************************************************************************
          S_LOAD_INST:                          cpu_a = pc;
//****************************************************************************
          S_EXECUTE_INST1:
            begin
              if(brk)                           cpu_a = V_IRQ;
              else                              cpu_a = { cpu_if, L };
            end
//****************************************************************************
          S_EXECUTE_INST2:
            begin
              if(branch_taken)                  cpu_a = branch;
              else                              cpu_a = { cpu_if, L };
            end
//****************************************************************************
          S_EXECUTE_INST3:
            begin
              if(st_axy|mem_op)                 cpu_a = { cpu_if, L };
              else if(ax)                       cpu_a = addx;
              else if(ay)                       cpu_a = addy;
              else                              cpu_a = add;
            end
//****************************************************************************
          S_RD_MEM_OP,S_RD_MEM_OP2,S_WR_MEM_OP: cpu_a = tmp_a;
//****************************************************************************
          S_INT: 
            if(~nmi)                            cpu_a = V_NMI;
            else                                cpu_a = V_IRQ;
//****************************************************************************
          default:                              cpu_a = { cpu_if, L };
        endcase
      end
    else if(~dma_ack)                           cpu_a = dma_a;
    else                                        cpu_a = { vic_a, L };

//****************************************************************************
//****************************************************************************
//read instruction, pc vector address bus
    if(~cpu_ack)
      begin
        case(state)
//****************************************************************************
          S_RESET:                              cpu_al = 14'h3FFF;
//****************************************************************************
          S_LOAD_INST:                          cpu_al = pc_rl;
//****************************************************************************
          S_EXECUTE_INST1:
            begin
              if(brk)                           cpu_al = 14'h3FFF;
              else                              cpu_al = ifl;
            end
//****************************************************************************
          S_EXECUTE_INST2:
            begin
              if(branch_taken)                  cpu_al = branchl;
              else                              cpu_al = ifl;
            end
//****************************************************************************
          S_EXECUTE_INST3:
            begin
              if(st_axy|mem_op)                 cpu_al = ifl;
              else if(ax)                       cpu_al = addxl;
              else if(ay)                       cpu_al = addyl;
              else                              cpu_al = addl;
            end
//****************************************************************************
          S_RD_MEM_OP,S_RD_MEM_OP2,S_WR_MEM_OP: cpu_al = tmp_a[15:2];
//****************************************************************************
          S_INT:                                cpu_al = 14'h3FFF;
//****************************************************************************
          default:                              cpu_al = ifl;
        endcase
      end
    else if(~dma_ack)                           cpu_al = dma_a[15:2];
    else                                        cpu_al = vic_a[15:2];

//****************************************************************************
//****************************************************************************
//next x look-ahead
    case(state)
//****************************************************************************
      S_EXECUTE_INST1:
        begin
          if(tax)      next_x = a;
          else if(tsx) next_x = sp;
          else if(dex) next_x = x - ONE;
          else if(inx) next_x = x + ONE;
          else         next_x = x;
        end
//****************************************************************************
      S_EXECUTE_INST2:
        begin
          if(ldx_i)    next_x = ir[15:8];
          else         next_x = x;
        end
//****************************************************************************
      S_REG_OP:
        if(ldx)        next_x = alu_op;
        else           next_x = x;
//****************************************************************************
      default:         next_x = x;
    endcase

//****************************************************************************
//****************************************************************************
//next y look-ahead
    case(state)
//****************************************************************************
      S_EXECUTE_INST1:
        begin
          if(tay)      next_y = a;
          else if(dey) next_y = y - ONE;
          else if(iny) next_y = y + ONE;
          else         next_y = y;
        end
//****************************************************************************
      S_EXECUTE_INST2:
        begin
          if(ldy_i)    next_y = ir[15:8];
          else         next_y = y;
        end
//****************************************************************************
      S_REG_OP:
        if(ldy)        next_y = alu_op;
        else           next_y = y;
//****************************************************************************
      default:         next_y = y;
    endcase

//****************************************************************************
//****************************************************************************
//next state look-ahead
    case(state)
//****************************************************************************
      S_RESET,S_INT:                              next_state = S_LOAD_INST;  
//****************************************************************************
      S_LOAD_INST:                                next_state = S_DECODE_INST;
//****************************************************************************
      S_DECODE_INST:
        if(one_byte_a)                            next_state = S_EXECUTE_INST1;
        else if(three_byte_a)                     next_state = S_EXECUTE_INST3;
        else                                      next_state = S_EXECUTE_INST2;
//****************************************************************************
      S_EXECUTE_INST1:
        if(rts)                                   next_state = S_RTS;
        else if(pla)                              next_state = S_REG_OP;
        else if(~nmi | ~rirq)                     next_state = S_INT;
        else if(brk|rti)                          next_state = S_LOAD_INST;
        else if(one_byte_a & two_valid)           next_state = S_EXECUTE_INST1;
        else if(three_byte_a & four_valid)        next_state = S_EXECUTE_INST3;
        else if(two_byte_a & three_valid)         next_state = S_EXECUTE_INST2;
        else                                      next_state = S_IR_LOAD;
//****************************************************************************
      S_EXECUTE_INST2:
        if(sta_ix|mem_op)                         next_state = S_WR_MEM_OP;
        else if(ix)                               next_state = S_RD_MEM_OP;
        else if(iy)                               next_state = S_IND_Y;
        else if(imm_addsub & decimal)             next_state = S_BCD_DELAY;
        else if(reg_op)                           next_state = S_REG_OP;
        else if(st_zxy)                           next_state = S_WR_MEM_OP;
        else if(~nmi | ~rirq)                     next_state = S_INT;
        else if(branch_taken)                     next_state = S_DECODE_INST;
        else if(one_byte_a & three_valid)         next_state = S_EXECUTE_INST1;
        else if(three_byte_a & five_valid)        next_state = S_EXECUTE_INST3;
        else if(two_byte_a & four_valid)          next_state = S_EXECUTE_INST2;
        else                                      next_state = S_IR_LOAD;
//****************************************************************************
      S_EXECUTE_INST3:
        if(st_axy)                                next_state = S_WR_MEM_OP;
        else if(mem_op)                           next_state = S_RD_MEM_OP2;
        else if(reg_op)                           next_state = S_REG_OP;
        else if(~nmi | ~rirq)                     next_state = S_INT;
        else if(jsr_a|jmp_a)                      next_state = S_DECODE_INST; 
        else /*if(jmp_i)*/                        next_state = S_LOAD_INST; 
//****************************************************************************
      S_RD_MEM_OP:                                next_state = S_REG_OP; 
//****************************************************************************
      S_RD_MEM_OP2:                               next_state = S_WR_MEM_OP;
//****************************************************************************
      S_IND_Y:  // indirect y
        if(sta_iy_2)                              next_state = S_WR_MEM_OP;
        else                                      next_state = S_RD_MEM_OP;
//****************************************************************************
      S_WR_MEM_OP:
        if(~nmi | ~rirq)                          next_state = S_INT;
        else if(cpu_a == pc)                      next_state = S_IR_HIT;
        else if((cpu_a == pc_p1r) & ~one_byte_a)  next_state = S_IR_HIT;
        else if((cpu_a == pc_p2r) & three_byte_a) next_state = S_IR_HIT;
        else if(one_byte_a & one_valid)           next_state = S_EXECUTE_INST1;
        else if(three_byte_a & three_valid)       next_state = S_EXECUTE_INST3;
        else if(two_byte_a & two_valid)           next_state = S_EXECUTE_INST2;
        else                                      next_state = S_IR_LOAD;
//****************************************************************************
      S_REG_OP:
        if(addsub_2 & decimal)                    next_state = S_BCD_DELAY;
        else if(~nmi | ~rirq)                     next_state = S_INT;
        else if(one_byte_a & one_valid)           next_state = S_EXECUTE_INST1;
        else if(three_byte_a & three_valid)       next_state = S_EXECUTE_INST3;
        else if(two_byte_a & two_valid)           next_state = S_EXECUTE_INST2;
        else                                      next_state = S_IR_LOAD;
//****************************************************************************
      S_BCD_DELAY,S_IR_LOAD,S_IR_HIT:
        if(~nmi | ~rirq)                          next_state = S_INT;
        else if(one_byte_a & one_valid)           next_state = S_EXECUTE_INST1;
        else if(three_byte_a & three_valid)       next_state = S_EXECUTE_INST3;
        else if(two_byte_a & two_valid)           next_state = S_EXECUTE_INST2;
        else                                      next_state = S_IR_LOAD;
//****************************************************************************
      S_RTS:
        if(~nmi | ~rirq)                          next_state = S_INT;
        else                                      next_state = S_LOAD_INST;
    endcase

//****************************************************************************
//****************************************************************************
//next pc look-ahead
    case(state)
//****************************************************************************
      S_RESET,S_INT:  //read the new program counter off the data bus
        begin
          case(cpu_a[1:0])
            0 :            next_pc = cpu_dar0;
            1 :            next_pc = cpu_dar1;
            2 :            next_pc = cpu_dar2;
            3 :            next_pc = cpu_dar3;
          endcase
        end
//****************************************************************************
      S_EXECUTE_INST1:
        begin
          if(rts)
            case(cpu_ia[1:0])
              0 :          next_pc = cpu_sar0;
              1 :          next_pc = cpu_sar1;
              2 :          next_pc = cpu_sar2;
              3 :          next_pc = cpu_sar3;
            endcase
          else if(rti)
            case(cpu_ia[1:0])
              0:           next_pc = cpu_sar1;
              1:           next_pc = cpu_sar2;
              2:           next_pc = cpu_sar3;
              3:           next_pc = cpu_sar0;
            endcase
          else             next_pc = pc_p1r;
        end
//****************************************************************************
      S_EXECUTE_INST2:
        begin
          if(branch_taken) next_pc = branch;
          else             next_pc = pc_p2r;
       end
//****************************************************************************
      S_EXECUTE_INST3:
        begin
          if(jsr_a|jmp_a)  next_pc = add;
          else if(jmp_i)
            begin
              case(cpu_ia[1:0])
                0 :        next_pc = (internal_hit & add[8]) ? cpu_sar0 : cpu_dar0;
                1 :        next_pc = (internal_hit & add[8]) ? cpu_sar1 : cpu_dar1;
                2 :        next_pc = (internal_hit & add[8]) ? cpu_sar2 : cpu_dar2;
                3 :        next_pc = (internal_hit & add[8]) ? cpu_sar3 : cpu_dar3;
              endcase
            end
          else             next_pc = pc_p3r;
        end
//****************************************************************************
      S_RTS:               next_pc = pc_p1r;
//****************************************************************************
      default:             next_pc = pc;
    endcase

//****************************************************************************
//****************************************************************************
//next if look-ahead
    case(state)
//****************************************************************************
      S_LOAD_INST:                                            next_if = pc[15:1] + 15'h0002;
//****************************************************************************
      S_EXECUTE_INST1:
        begin
          if(iCount < (IR_QUEUE_SIZE - 4'h2))                 next_if = cpu_ifp4;
          else                                                next_if = cpu_if;
        end
//****************************************************************************
      S_EXECUTE_INST2:
        begin
          if(branch_taken)                                    next_if = branch[15:1] + 15'h0002;
          else if(iCount < (IR_QUEUE_SIZE - 4'h1))            next_if = cpu_ifp4;
          else                                                next_if = cpu_if;
        end
//****************************************************************************
      S_EXECUTE_INST3:
        begin
          if(jsr_a|jmp_a)                                     next_if = add[15:1] + 15'h0002;
          else if((st_axy|mem_op) & (iCount < IR_QUEUE_SIZE)) next_if = cpu_ifp4;//delayed write/rmw  
          else                                                next_if = cpu_if;
        end
//****************************************************************************
      S_DECODE_INST,S_IR_LOAD,S_IND_Y,S_REG_OP,S_BCD_DELAY,S_IR_HIT:
        begin
          if(iCount < (IR_QUEUE_SIZE - 4'h3))                 next_if = cpu_ifp4;
          else                                                next_if = cpu_if;
        end
//****************************************************************************
      S_RTS:                                                  next_if = pc_p1r[15:1] + 15'h0002;
//****************************************************************************
      default:                                                next_if = cpu_if;
    endcase

//****************************************************************************
//****************************************************************************
//next if look-ahead
    case(state)
//****************************************************************************
      S_EXECUTE_INST1:
        begin
          if(brk)          next_sp = spm3;
          else if(rts)     next_sp = spp2;
          else if(rti)     next_sp = spp3;
          else if(txs)     next_sp = x;
          else if(plp|pla) next_sp = spp1;
          else if(pha|php) next_sp = spm1;
          else             next_sp = sp;
        end
//****************************************************************************
      S_EXECUTE_INST3:
        begin
          if(jsr_a)        next_sp = spm2;
          else             next_sp = sp;
       end
//****************************************************************************
      S_INT:               next_sp = spm3;
//****************************************************************************
      default:             next_sp = sp;
    endcase

//****************************************************************************
//****************************************************************************
//next C flag
    case(state)
//****************************************************************************
      S_EXECUTE_INST1:
        begin
          if(brk|rts)                   next_C = carry;
          else if(rti)
            begin
              case(spp1[1:0])
                0:                      next_C = internal_s[0];
                1:                      next_C = internal_s[8];
                2:                      next_C = internal_s[16];
                3:                      next_C = internal_s[24];
              endcase
            end
          else
            begin
              if(plp)
                begin
                  case(sp[1:0])
                    0:                  next_C = internal_s[8];
                    1:                  next_C = internal_s[16];
                    2:                  next_C = internal_s[24];
                    3:                  next_C = internal_s[0];
                  endcase
                end
              else if(asla | rola)      next_C = a[7];
              else if(lsra | rora)      next_C = a[0];
              else if(clc)              next_C = FALSE;
              else if(sec)              next_C = TRUE;
              else                      next_C = carry;
            end
        end
//****************************************************************************
      S_EXECUTE_INST2:
        begin
          if(~branch_taken & two_byte)
            begin
              if(imm_addsub & ~decimal) next_C = Cout_a;
              else if(cmpa_i)           next_C = Cout_ca;
              else if(cpy_i)            next_C = Cout_y;
              else if(cpx_i)            next_C = Cout_x;
              else                      next_C = carry;
            end
          else next_C = carry;
        end
//****************************************************************************
      S_WR_MEM_OP: // complete memory operation
        begin
          if(asl | rol)                 next_C = alu_op[7];
          else if(lsr | ror)            next_C = alu_op[0];
          else                          next_C = carry;
        end
//****************************************************************************
      S_REG_OP: // complete register operation
        begin
          if(cmpa_2)                    next_C = Cout_ca;
          else if(cpy_2)                next_C = Cout_y;
          else if(cpx_2)                next_C = Cout_x;
          else if(addsub_2 & ~decimal)  next_C = Cout_a;
          else                          next_C = carry;
        end
//****************************************************************************
      S_BCD_DELAY:                      next_C = bcdCout_a;
//****************************************************************************
      default:                          next_C = carry;
    endcase

//****************************************************************************
//****************************************************************************
//next V flag
    case(state)
//****************************************************************************
      S_EXECUTE_INST1:
        begin
          if(brk|rts)                   next_V = overflow;
          else if(rti)
            begin
              case(spp1[1:0])
                0:                      next_V = internal_s[6];
                1:                      next_V = internal_s[14];
                2:                      next_V = internal_s[22];
                3:                      next_V = internal_s[30];
              endcase
            end
          else
            begin
              if(plp)
                begin
                  case(sp[1:0])
                    0:                  next_V = internal_s[14];
                    1:                  next_V = internal_s[22];
                    2:                  next_V = internal_s[30];
                    3:                  next_V = internal_s[6];
                  endcase
                end
              else if(clv)              next_V = FALSE;
              else                      next_V = overflow;
            end
        end
//****************************************************************************
      S_EXECUTE_INST2:
        begin
          if(~branch_taken & two_byte)
            begin
              if(imm_addsub & ~decimal) next_V = is_overflow_a;
              else                      next_V = overflow;
            end
          else                          next_V = overflow;
        end
//****************************************************************************
      S_REG_OP: // complete register operation
        begin
          if(bita)                      next_V = alu_op[6];
          else if(addsub_2 & ~decimal)  next_V = is_overflow_a;
          else                          next_V = overflow;
        end
//****************************************************************************
      S_BCD_DELAY:                      next_V = bcdOverflow_a;
//****************************************************************************
      default:                          next_V = overflow;
    endcase

//****************************************************************************
//****************************************************************************
//next N flag
    case(state)
//****************************************************************************
      S_EXECUTE_INST1:
        begin
          if(brk|rts)                        next_N = negative;
          else if(rti)
            begin
              case(spp1[1:0])
                0:                           next_N = internal_s[7];
                1:                           next_N = internal_s[15];
                2:                           next_N = internal_s[23];
                3:                           next_N = internal_s[31];
              endcase
            end
          else
            begin
              if(plp)
                begin
                  case(sp[1:0])
                    0:                       next_N = internal_s[15];
                    1:                       next_N = internal_s[23];
                    2:                       next_N = internal_s[31];
                    3:                       next_N = internal_s[7];
                  endcase
                end
              else if(asla | rola)           next_N = a[6];
              else if(lsra)                  next_N = L;
              else if(rora)                  next_N = carry;
              else if(tya)                   next_N = y[7];
              else if(txa)                   next_N = x[7];
              else if(tay | tax)             next_N = a[7];
              else if(dey)                   next_N = (y == ZERO) ? H : (y == 8'h80) ? L : y[7];
              else if(iny)                   next_N = (y == NEG_ONE) ? L : (y == 8'h7F) ? H : y[7];
              else if(tsx)                   next_N = sp[7];
              else if(dex)                   next_N = (x == ZERO) ? H : (x == 8'h80) ? L : x[7];
              else if(inx)                   next_N = (x == NEG_ONE) ? L : (x == 8'h7F) ? H : x[7];
              else                           next_N = negative;
            end
        end
//****************************************************************************
      S_EXECUTE_INST2:
        begin
          if(~branch_taken & two_byte)
            begin
              if(anda_i)                     next_N = and_op[7];
              else if(eora_i)                next_N = eor_op[7];
              else if(ora_i)                 next_N = or_op[7];
              else if(lda_i | ldy_i | ldx_i) next_N = alu_op[7];
              else if(imm_addsub & ~decimal) next_N = is_negative_a;
              else if(cmpa_i)                next_N = N_ca;
              else if(cpy_i)                 next_N = N_y;
              else if(cpx_i)                 next_N = N_x;
              else                           next_N = negative;
            end
          else                               next_N = negative;
        end
//****************************************************************************
      S_WR_MEM_OP: // complete memory operation
        begin
          if(asl | rol)                      next_N = alu_op[6];
          else if(lsr)                       next_N = L;
          else if(ror)                       next_N = carry;
          else if(dec)                       next_N = alu_op_dec_neg;
          else if(inc)                       next_N = alu_op_inc_neg;
          else                               next_N = negative; //ix/iy writes
        end
//****************************************************************************
      S_REG_OP: // complete register operation
        begin
          if(ora)                            next_N = or_op[7];
          else if(bita | ldx | lda | ldy)    next_N = alu_op[7];
          else if(anda)                      next_N = and_op[7];
          else if(eora)                      next_N = eor_op[7];
          else if(cmpa_2)                    next_N = N_ca;
          else if(cpy_2)                     next_N = N_y;
          else if(cpx_2)                     next_N = N_x;
          else if(addsub_2 & ~decimal)       next_N = is_negative_a;
          else                               next_N = negative;
        end
//****************************************************************************
      S_BCD_DELAY:                           next_N = bcdNegative_a;
//****************************************************************************
      default:                               next_N = negative;
    endcase

//****************************************************************************
//****************************************************************************
//next Z flag
    case(state)
//****************************************************************************
      S_EXECUTE_INST1:
        begin
          if(brk|rts)                        next_Z = zero;
          else if(rti)
            begin
              case(spp1[1:0])
                0:                           next_Z = internal_s[1];
                1:                           next_Z = internal_s[9];
                2:                           next_Z = internal_s[17];
                3:                           next_Z = internal_s[25];
              endcase
            end
          else
            begin
              if(plp)
                begin
                  case(sp[1:0])
                    0:                       next_Z = internal_s[9];
                    1:                       next_Z = internal_s[17];
                    2:                       next_Z = internal_s[25];
                    3:                       next_Z = internal_s[1];
                  endcase
                end
              else if(asla)                  next_Z = ~(|a[6:0]);
              else if(rola)                  next_Z = ~carry & ~(|a[6:0]);
              else if(lsra)                  next_Z = ~(|a[7:1]);
              else if(rora)                  next_Z = ~carry & ~(|a[7:1]);
              else if(tya)                   next_Z = (y == ZERO);
              else if(txa)                   next_Z = (x == ZERO);
              else if(tay)                   next_Z = (a == ZERO);
              else if(dey)                   next_Z = (y == ONE);
              else if(iny)                   next_Z = (y == NEG_ONE);
              else if(tax)                   next_Z = (a == ZERO);
              else if(tsx)                   next_Z = (sp == ZERO);
              else if(dex)                   next_Z = (x == ONE);
              else if(inx)                   next_Z = (x == NEG_ONE);
              else                           next_Z = zero;
            end
        end
//****************************************************************************
      S_EXECUTE_INST2:
        begin
          if(~branch_taken & two_byte)
            begin
              if(anda_i)                     next_Z = and_op_zero;
              else if(eora_i)                next_Z = eor_op_zero;
              else if(ora_i)                 next_Z = or_op_zero;
              else if(lda_i | ldy_i | ldx_i) next_Z = alu_op_0;
              else if(imm_addsub & ~decimal) next_Z = is_zero_a;
              else if(cmpa_i)                next_Z = is_zero_ca;
              else if(cpy_i)                 next_Z = is_zero_y;
              else if(cpx_i)                 next_Z = is_zero_x;
              else                           next_Z = zero;
            end
          else                               next_Z = zero;
        end
//****************************************************************************
      S_WR_MEM_OP: // complete memory operation
        begin
          if(asl)                            next_Z = ~(|alu_op[6:0]);
          else if(rol)                       next_Z = ~carry & ~(|alu_op[6:0]);
          else if(lsr)                       next_Z = ~(|alu_op[7:1]);
          else if(ror)                       next_Z = ~carry & ~(|alu_op[7:1]);
          else if(dec)                       next_Z = alu_op_1;
          else if(inc)                       next_Z = alu_op_255;
          else                               next_Z = zero; //ix/iy writes
        end
//****************************************************************************
      S_REG_OP: // complete register operation
        begin
          if(ora)                            next_Z = or_op_zero;
          else if(bita | anda)               next_Z = and_op_zero;
          else if(eora)                      next_Z = eor_op_zero;
          else if(cmpa_2)                    next_Z = is_zero_ca;
          else if(cpy_2)                     next_Z = is_zero_y;
          else if(cpx_2)                     next_Z = is_zero_x;
          else if(ldx | lda | ldy)           next_Z = alu_op_0;
          else if(addsub_2 & ~decimal)       next_Z = is_zero_a;
          else                               next_Z = zero;
        end
//****************************************************************************
      S_BCD_DELAY:                           next_Z = bcdZero_a;
//****************************************************************************
      default:                               next_Z = zero;
    endcase
  end


`ifdef debug
//debug trace registers
reg [959:0] pc_trace = 960'd0;
reg [959:0] ad_trace = 960'd0;
reg [959:0] aw_trace = 960'd0;

//for pausing cpu due to certain events - starts single step
reg breakTrig = FALSE;
reg breakTrig1 = FALSE;

reg [31:0] cpu_rdr, internal_rr;
reg [15:0] cpu_a2_i, cpu_sa2_i;
reg [15:2] cpu_al_i;
reg [7:0] cpu_pr_r, cpu_wb_r;
reg [3:0] cpu_byte_enable, cpu_byte_enablez;
reg cpu_rw_r, cpu_rwz_r;
 
//for counting events
reg [31:0] cycleCount, cycleCount2;
reg enableCycleCounts;
reg disableI = FALSE;
`endif

//synchronous cpu state machine
//****************************************************************************
//****************************************************************************
//instruction execution state machine controller
always@(posedge clk)
  begin
    //bus master logic
    vic_ack <= vic_req;
    dma_ack <= ~vic_req|dma_req|(~dma_ack & RDY); //must only be a one cycle steal
    cpu_ack <= ~(vic_req & ~(vic_req & dma_ack & ~dma_req) & cpu_pause & resetIn & (&halt));    
       
    //cpu is being reset
    if(~resetIn)
      begin
        halt <= 4'hF; 

        rnmi <= H;
        rnmi2 <= H; //this needs to be cleared to allow nmi's
        nmi <= H;
        rirq <= H;
        
        state <= S_RESET;
        pc <= ~dma_a;

//the rest of these are optional
        iCount <= 4'h0;

`ifdef debug
        pc_trace <= 960'd0;
        ad_trace <= 960'd0;
        aw_trace <= 960'd0;
        breakTrig <= FALSE;
        breakTrig1 <= FALSE;
        cycleCount <= 0; //count events
        cycleCount2 <= 0; //count events
        enableCycleCounts <= TRUE;
        disableI <= FALSE;
`endif
      end
    else if(~cpu_ack & RDY) //cpu is the master and the current memory access is complete
      begin
        state <= next_state;
        
        x <= next_x;
        y <= next_y;
        
        carry <= next_C;
        overflow <= next_V;
        negative <= next_N;
        zero <= next_Z;

        addx <= add_x;
        addy <= add_y;

        pc <= next_pc;
        pc_p1r <= pcp1;
        pc_p2r <= pcp2;
        pc_p3r <= pcp3;
        pc_p4r <= pcp4;
        pc_p5r <= pcp5;
        pc_p6r <= pcp6;
        pc_p7r <= pcp7;
        pc_p8r <= pcp8;

        cpu_if <= next_if;
        
        sp <= next_sp;
        spp1 <= next_spp1;
        spp2 <= next_spp2;
        spp3 <= next_spp3;
        
        //set up the branch address assuming the next instrution is going to be a branch
        if(state == S_EXECUTE_INST1)      branch <= branchp1;
        else if(state == S_EXECUTE_INST2) branch <= branchp2;
        else                              branch <= branchp0;

        // for cpu_al
        if(state == S_EXECUTE_INST1)      branchl <= branchp1[15:2] + branchp1[1];
        else if(state == S_EXECUTE_INST2) branchl <= branchp2[15:2] + branchp2[1];
        else                              branchl <= branchp0[15:2] + branchp0[1];
        
        addl <= { ah, imm_op[7:2] } + imm_op[1];
        addxl <= add_x[15:2] + add_x[1];
        addyl <= add_y[15:2] + add_y[1];
        pc_rl <= next_pc[15:2] + next_pc[1];
        ifl <= next_if[15:2] + next_if[1];
        
        //load #, zp internal operand or memory operand
        if(next_state == S_EXECUTE_INST2)       alu_op <= imm_op;
        else if(next_state == S_BCD_DELAY)      alu_op <= alu_op;
        else if((state == S_EXECUTE_INST1) |                //for pla
                (state == S_EXECUTE_INST2) |                //for zp operands
                (internal_hit & ((state == S_EXECUTE_INST3) |
                                 (state == S_RD_MEM_OP)     |
                                 (state == S_RD_MEM_OP2)))) //for absoluted addressing that falls into page 0 or 1
          begin
            case(cpu_ia[1:0])
              0:                                alu_op <= internal[7:0];
              1:                                alu_op <= internal[15:8];
              2:                                alu_op <= internal[23:16];
              3:                                alu_op <= internal[31:24];
            endcase
          end
        else if(pRead)                          alu_op <= cpu_pr;
        else
          case(cpu_a[1:0])
            0 :                                 alu_op <= cpu_rd[7:0];
            1 :                                 alu_op <= cpu_rd[15:8];
            2 :                                 alu_op <= cpu_rd[23:16];
            3 :                                 alu_op <= cpu_rd[31:24];
          endcase
   
        if((next_state == S_EXECUTE_INST2) |
           (next_state == S_EXECUTE_INST3))     alu_mode <= ira[7];
        
//remember operations for later states
        if((state == S_EXECUTE_INST1)|(state == S_EXECUTE_INST2)|(state == S_EXECUTE_INST3))
          begin
            sta_iy_2 <= sta_iy;
            stx_2 <= stx_z|stx_zy|stx_a;
            sty_2 <= sty_z|sty_zx|sty_a;
              
            asl <= asl_z|asl_zx|asl_a|asl_ax;
            rol <= rol_z|rol_zx|rol_a|rol_ax;
            lsr <= lsr_z|lsr_zx|lsr_a|lsr_ax;
            ror <= ror_z|ror_zx|ror_a|ror_ax;
            dec <= dec_z|dec_zx|dec_a|dec_ax;
            inc <= inc_z|inc_zx|inc_a|inc_ax;

            ora <= ora_iy|ora_ix|ora_z|ora_zx|ora_a|ora_ay|ora_ax;
            bita <= bita_z|bita_a;
            anda <= anda_iy|anda_ix|anda_z|anda_zx|anda_a|anda_ay|anda_ax;
            eora <= eora_iy|eora_ix|eora_z|eora_zx|eora_a|eora_ay|eora_ax;
              
            addsub_2 <= addsub;
              
            cmpa_2 <= cmpa_iy|cmpa_ix|cmpa_z|cmpa_zx|cmpa_z|cmpa_zx|cmpa_a|cmpa_ay|cmpa_ax;
            cpx_2 <= cpx_z|cpx_a;
            cpy_2 <= cpy_z|cpy_a;
              
            ldx <= ldx_zy|ldx_z|ldx_a|ldx_ay;
            lda <= lda_iy|lda_ix|lda_z|lda_zx|lda_a|lda_ay|lda_ax|pla;
            ldy <= ldy_z|ldy_zx|ldy_a|ldy_ax;
          end

        //update interrupts
        rnmi <= nmi_req;
        rnmi2 <= rnmi;
`ifdef debug
        rirq <= irq | interrupt | disableI;
`else
        rirq <= irq | interrupt;
`endif
        if(~rnmi & rnmi2) nmi <= L; //edge trigger until processed

`ifdef debug

        ad_trace[15:0] <= cpu_a;
        ad_trace[959:16] <= ad_trace[943:0];
        aw_trace[15:0] <= cpu_ia;
        aw_trace[959:16] <= aw_trace[943:0];

        breakTrig <= breakTrig1;

        cpu_a2_i <= cpu_a;
        cpu_al_i <= cpu_al;
        cpu_rw_r <= cpu_rw;
        cpu_byte_enable <= cpu_wbyte;

        cpu_sa2_i <= cpu_ia;
        cpu_rwz_r <= cpu_rwz;
        cpu_byte_enablez <= cpu_wbytez;
        internal_rr <= internal_s;

        cpu_rdr <= cpu_rd;
        cpu_pr_r <= cpu_pr;
        cpu_wb_r <= cpu_wb;

        //count events
        if(pc == dma_a)
          begin
            enableCycleCounts <= TRUE;
            cycleCount <= 0; //count events
            cycleCount2 <= 0; //count events
            disableI <= TRUE;
          end
        //this is a loop at boot time that delays the process
        //if vic vcounter is not changing & is not zero the code will stop  here
        else if(pc == 16'hFF5E) enableCycleCounts <= FALSE;
        else if(pc == 16'hFF63) enableCycleCounts <= TRUE;
        else if(state == S_INT) enableCycleCounts <= FALSE;//(break_pc == 16'h37B9);
        else if(pc == break_pc)
          begin
            enableCycleCounts <= FALSE;
            disableI <= FALSE;
          end

`ifdef debug_cycles
        if(enableCycleCounts)
          begin
            cycleCount <= cycleCount + H; //count events
            if(cycleCount == 32'hFFFFFFFF) cycleCount2 <= cycleCount2 + H; //count events
          end
`endif

`endif
        case(state)
//****************************************************************************
          S_RESET:
            begin
              one_valid <= FALSE;
              two_valid <= FALSE;
              three_valid <= FALSE;
              four_valid <= FALSE;
              five_valid <= FALSE;
              ir[IR_QUEUE_SIZE*8-1:0] <= 0;
              iCount <= 4'h0;
            end
//****************************************************************************
//transfer the first instruction from the new path to cpu_dr & start accessing
//the next 4 bytes in the instruction stream
          S_LOAD_INST:
            begin
              one_valid <= TRUE;
              two_valid <= TRUE;
              three_valid <= TRUE;
              four_valid <= pc[0] ? FALSE : TRUE;
              five_valid <= FALSE;
              ir[IR_QUEUE_SIZE*8-1:32] <= 0;
              iCount <= pc[0] ? 4'h3 : 4'h4;
              case(pc[1:0])
                0 : ir[31:0] <= cpu_rd[31:0];
                1 : ir[31:0] <= { ZERO, cpu_rd[31:8] }; 
                2 : ir[31:0] <= { cpu_rd[15:0], cpu_rd[31:16] }; 
                3 : ir[31:0] <= { ZERO, cpu_rd[15:0], cpu_rd[31:24] }; 
              endcase
            end

//****************************************************************************
//transfer the next instruction bytes from the new path to cpu_rd & start accessing
//the next 4 bytes in the instruction stream
          S_DECODE_INST:
            begin
              one_valid <= TRUE;
              two_valid <= TRUE;
              three_valid <= TRUE;
              four_valid <= TRUE;
              five_valid <= TRUE;
              case(iCount[0])
                L :
                  begin
                    ir[63:32] <= ib32;
                  end
                H :
                  begin
                    ir[55:24] <= ib32;
                  end
              endcase
              iCount <= iCount + 4'h4;
            end
            
//****************************************************************************
//start executing a single byte instruction
          S_EXECUTE_INST1:
            begin
`ifdef debug
              breakTrig1 <= (pc == break_pc);
              pc_trace[15:0] <= pc;
              pc_trace[959:16] <= pc_trace[943:0];
`endif
              if(brk)
                begin
//brk is same as irq except for break status bit set on the stack and pc + 2
//is stacked, note this is a software interrupt, uses irq vector
                  interrupt <= H;
                end
              else if(rts)
                begin
                  iCount <= 4'h0;//do nothing
                end
              else if(rti)
                begin
                  case(spp1[1:0])
                    0: interrupt <= internal_s[2];
                    1: interrupt <= internal_s[10];
                    2: interrupt <= internal_s[18];
                    3: interrupt <= internal_s[26];
                  endcase
                  case(spp1[1:0])
                    0: decimal <= internal_s[3];
                    1: decimal <= internal_s[11];
                    2: decimal <= internal_s[19];
                    3: decimal <= internal_s[27];
                  endcase
                end
              else
                begin
                  if(iCount < (IR_QUEUE_SIZE - 4'h2))
                    begin
                      one_valid <= TRUE;
                      two_valid <= TRUE;
                      three_valid <= TRUE;
                      four_valid <= (iCount > 0);
                      five_valid <= (iCount > 1);
                      case(iCount)
                        6: ir[71:0]  <= { ib32, ir[47:8] };
                        5: ir[63:0]  <= { ib32, ir[39:8] };
                        4: ir[55:0]  <= { ib32, ir[31:8] };
                        3: ir[47:0]  <= { ib32, ir[23:8] };
                        2: ir[39:0]  <= { ib32, ir[15:8] };
                        1: ir[31:0]  <=   ib32;
                        default: halt <= 4'h0; //this should not happen... 
                      endcase
                      //subtracting 1 and adding 4
                      iCount <= iCount + 4'h3;
                    end
                  else
                    begin
                      one_valid <= (iCount > 1);
                      two_valid <= (iCount > 2);
                      three_valid <= (iCount > 3);
                      four_valid <= (iCount > 4);
                      five_valid <= (iCount > 5);
                      ir[IR_QUEUE_SIZE*8-1:0] <= { 8'h00, ir[IR_QUEUE_SIZE*8-1:8] };
                      iCount <= iCount - 4'h1;
                    end
                
                  if(plp)
                    begin
                      case(sp[1:0])
                        0: interrupt <= internal_s[10];
                        1: interrupt <= internal_s[18];
                        2: interrupt <= internal_s[26];
                        3: interrupt <= internal_s[2];
                      endcase
                      case(sp[1:0])
                        0: decimal <= internal_s[11];
                        1: decimal <= internal_s[19];
                        2: decimal <= internal_s[27];
                        3: decimal <= internal_s[3];
                      endcase
                    end
                    
                  if(asla) a <= { a[6:0], L };
                  if(rola) a <= { a[6:0], carry };
                  if(lsra) a <= { L, a[7:1] };
                  if(rora) a <= { carry, a[7:1] };
                  if(tya)  a <= y;
                  if(txa)  a <= x;
                  if(cli) interrupt <= FALSE;
                  if(sei) interrupt <= TRUE;
                  if(cld) decimal <= FALSE;
                  if(sed) decimal <= TRUE;
                end
            end
//****************************************************************************
//start executing a 2 byte instruction
          S_EXECUTE_INST2:
            begin
`ifdef debug
              breakTrig1 <= (pc == break_pc);
              pc_trace[15:0] <= pc;
              pc_trace[959:16] <= pc_trace[943:0];
`endif
              if(branch_taken)
                begin
                  one_valid <= TRUE;
                  two_valid <= TRUE;
                  three_valid <= TRUE;
                  four_valid <= branch[0] ? FALSE : TRUE;
                  five_valid <= FALSE;
                  ir[IR_QUEUE_SIZE*8-1:32] <= 0;
                  iCount <= branch[0] ? 4'h3 : 4'h4;
                  case(branch[1:0])
                    0 : ir[31:0] <= cpu_rd[31:0];
                    1 : ir[31:0] <= { ZERO, cpu_rd[31:8] }; 
                    2 : ir[31:0] <= { cpu_rd[15:0], cpu_rd[31:16] }; 
                    3 : ir[31:0] <= { ZERO, cpu_rd[15:0], cpu_rd[31:24] }; 
                  endcase
                end
              else if(two_byte) //necessary to catch illegal instructions below
                begin
                  if(iCount < (IR_QUEUE_SIZE - 4'h1))
                    begin
                      one_valid <= TRUE;
                      two_valid <= TRUE;
                      three_valid <= (iCount > 0);
                      four_valid <= (iCount > 1);
                      five_valid <= (iCount > 2);
                      case(iCount)
                        7: ir[71:0]  <= { ib32, ir[55:16] };
                        6: ir[63:0]  <= { ib32, ir[47:16] };
                        5: ir[55:0]  <= { ib32, ir[39:16] };
                        4: ir[47:0]  <= { ib32, ir[31:16] };
                        3: ir[39:0]  <= { ib32, ir[23:16] };
                        2: ir[31:0]  <=   ib32;
                        default: //too many bytes already in the queue
                          begin
                            one_valid <= (iCount > 2);
                            two_valid <= (iCount > 3);
                            three_valid <= (iCount > 4);
                            four_valid <= (iCount > 5);
                            five_valid <= (iCount > 6);
                            //six_valid <= (iCount > 7);
                            iCount <= iCount - 4'h2;
                            ir[IR_QUEUE_SIZE*8-1:0] <= { 16'h0000, ir[IR_QUEUE_SIZE*8-1:16] };
                          end
                      endcase
                      //subtracting 2 and adding 4
                      iCount <= iCount + 4'h2;
                    end
                  else
                    begin
                      one_valid <= (iCount > 2);
                      two_valid <= (iCount > 3);
                      three_valid <= (iCount > 4);
                      four_valid <= (iCount > 5);
                      five_valid <= (iCount > 6);
                      ir[IR_QUEUE_SIZE*8-1:0] <= { 16'h0000, ir[IR_QUEUE_SIZE*8-1:16] };
                      iCount <= iCount - 4'h2;
                    end
                  if(anda_i)                a <= and_op;
                  if(eora_i)                a <= eor_op;
                  if(ora_i)                 a <= or_op;
                  if(lda_i)                 a <= alu_op;
                  if(imm_addsub & ~decimal) a <= result;

                  if(zx)      tmp_a <= { ZERO, addx[7:0] };
                  else if(zy) tmp_a <= { ZERO, addy[7:0] };
                  else if(ix)
                    begin
                      case(addx[1:0])
                        0 :   tmp_a <= cpu_zar0;
                        1 :   tmp_a <= cpu_zar1;
                        2 :   tmp_a <= cpu_zar2;
                        3 :   tmp_a <= cpu_zar3;
                      endcase
                    end
                  else if(iy)
                    begin
                      case(add[1:0])
                        0 :   tmp_a <= cpu_zar0;
                        1 :   tmp_a <= cpu_zar1;
                        2 :   tmp_a <= cpu_zar2;
                        3 :   tmp_a <= cpu_zar3;
                      endcase
                    end
                  else        tmp_a <= { ZERO, add[7:0] };
                end
//****************************************************************************
//illegal instructions
              else halt <= 4'h1;
            end
//****************************************************************************
//start executing a 3 byte instruction
          S_EXECUTE_INST3:
            begin
`ifdef debug
              breakTrig1 <= (pc == break_pc);
              pc_trace[15:0] <= pc;
              pc_trace[959:16] <= pc_trace[943:0];
`endif
              if(jsr_a|jmp_a)
                begin
                  one_valid <= TRUE;
                  two_valid <= TRUE;
                  three_valid <= TRUE;
                  four_valid <= add[0] ? FALSE : TRUE;
                  five_valid <= FALSE;
                  ir[IR_QUEUE_SIZE*8-1:32] <= 0;
                  case(add[1:0])
                    0 : ir[31:0] <= cpu_rd[31:0];
                    1 : ir[31:0] <= { ZERO, cpu_rd[31:8] }; 
                    2 : ir[31:0] <= { cpu_rd[15:0], cpu_rd[31:16] }; 
                    3 : ir[31:0] <= { ZERO, cpu_rd[15:0], cpu_rd[31:24] }; 
                  endcase
                  iCount <= add[0] ? 4'h3 : 4'h4;
                end
              else
                begin
                  if((st_axy|mem_op) & (iCount < IR_QUEUE_SIZE))
                    begin
                      one_valid <= TRUE;
                      two_valid <= (iCount > 0);
                      three_valid <= (iCount > 1);
                      four_valid <= (iCount > 2);
                      five_valid <= (iCount > 3);
                      case(iCount)
                        8: ir[71:0] <= { ib32, ir[63:24] };
                        7: ir[63:0] <= { ib32, ir[55:24] };
                        6: ir[55:0] <= { ib32, ir[47:24] };
                        5: ir[47:0] <= { ib32, ir[39:24] };
                        4: ir[39:0] <= { ib32, ir[31:24] };
                        3: ir[31:0] <= ib32;
                        default: halt <= 4'h2; //should not happen
                      endcase
                      //subtracting 3 and adding 4
                      iCount <= iCount + 4'h1;
                    end
                  else
                    begin
                      one_valid <= (iCount > 3);
                      two_valid <= (iCount > 4);
                      three_valid <= (iCount > 5);
                      four_valid <= (iCount > 6);
                      five_valid <= (iCount > 7);
                      iCount <= iCount - 4'h3;
                      ir[IR_QUEUE_SIZE*8-1:0] <= { 24'h000000, ir[IR_QUEUE_SIZE*8-1:24] };
                      iCount <= iCount - 4'h3;
                    end
                    
                  if(ax)      tmp_a <= addx;
                  else if(ay) tmp_a <= addy; //no mem-mem ops with this addressing mode
                  else        tmp_a <= add;
                end
            end
//****************************************************************************
          S_IND_Y,S_REG_OP,S_BCD_DELAY,S_IR_LOAD,S_IR_HIT:
            begin
              if(iCount < (IR_QUEUE_SIZE - 4'h3))
                begin
                  one_valid <= TRUE;
                  two_valid <= TRUE;
                  three_valid <= TRUE;
                  four_valid <= TRUE;
                  five_valid <= (iCount > 0);
                  case(iCount)
                    5: ir[71:40] <= ib32;
                    4: ir[63:32] <= ib32;
                    3: ir[55:24] <= ib32;
                    2: ir[47:16] <= ib32;
                    1: ir[39:8]  <= ib32;
                    0: ir[31:0]  <= ib32;
                    default: halt <= 4'h3; //this should not happen
                  endcase
                  iCount <= iCount + 4'h4;
                end

              case(state)
                S_IND_Y:                    tmp_a <= tmp_a + y;
                S_REG_OP:
                  begin
                    if(ora)                 a <= or_op;
                    if(anda)                a <= and_op;
                    if(eora)                a <= eor_op;
                    if(lda)                 a <= alu_op;
                    if(addsub_2 & ~decimal) a <= result;
                  end
                S_BCD_DELAY:                a <= bcdResult;
`ifdef debug
`ifndef debug_cycles
                S_IR_LOAD: if(enableCycleCounts) cycleCount2 <= cycleCount2 + H;
                S_IR_HIT:  if(enableCycleCounts) cycleCount <= cycleCount + H;
`endif
`endif
              endcase
            end
//****************************************************************************
          S_WR_MEM_OP: //write state - update ir queue if necesary
            begin
              if(cpu_a == pc_p8r) ir[71:64] <= cpu_wb;
              if(cpu_a == pc_p7r) ir[63:56] <= cpu_wb;
              if(cpu_a == pc_p6r) ir[55:48] <= cpu_wb;
              if(cpu_a == pc_p5r) ir[47:40] <= cpu_wb;
              if(cpu_a == pc_p4r) ir[39:32] <= cpu_wb;
              if(cpu_a == pc_p3r) ir[31:24] <= cpu_wb;
              if(cpu_a == pc_p2r) ir[23:16] <= cpu_wb;
              if(cpu_a == pc_p1r) ir[15:8]  <= cpu_wb;
              if(cpu_a == pc)     ir[7:0]   <= cpu_wb;
            end
//****************************************************************************
//start processing an nmi or irq
          S_INT:
            begin
              if(~nmi) nmi <= TRUE;
//according to some documentation this else should not be here...
              /*else*/ interrupt <= H;
            end
        endcase
    end
end

`ifdef debug
cpu_debug dbg(clk, pc_trace, ad_trace, aw_trace
             ,state,halt,cpu_rwz_r,cpu_byte_enablez,internal_rr,cpu_rw_r,cpu_byte_enable,a1
             ,negative,decimal,interrupt,overflow,zero,carry,rirq,nmi
             ,{ZERO,ZERO,ZERO,ir[IR_QUEUE_SIZE*8-1:0]},iCount
             ,a,x,y,sp,alu_op,cpu_pr_r,cpu_wb_r
             ,pc,tmp_a,cpu_if,cpu_a2_i,cpu_al_i,cpu_sa2_i,break_pc
             ,cpu_rdr, cycleCount, cycleCount2
             ,debug_add, debug_Dout, debug_sel
             );
`endif

endmodule


